///
/// @package phosim
/// @file trim.cpp
/// @brief trim program: removes sources that have no chance of producing photons on a chip
///
/// @brief Created by:
/// @author Alan Meert (Purdue)
///
/// @brief Modified by:
/// @author Justin Bankert (Purdue)
/// @author John R. Peterson (Purdue)
/// @author En-Hsin Peng (Purdue)
///
/// @warning This code is not fully validated
/// and not ready for full release.  Please
/// treat results with caution.
///

#include <stdio.h>
#include <zlib.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>

#include "trim/trim.h"
#include "raytrace/constants.h"
#include "ancillary/readtext.h"

using readtext::readText;

void Trim::xyPosition(double alpha, double delta, double phiChip, double psiChip, double thetaChip, double phiImage, double psiImage, double thetaImage, double xPos, double yPos, double deltaX, double deltaY, double deltaZ, double *x, double *y) {
    
    // find x, y position at focus
    double a = cos(delta)*cos(alpha - pointingRA);
    double f = focalLength/(sin(pointingDec)*sin(delta) + a*cos(pointingDec));
    double yp = f*(cos(pointingDec)*sin(delta) - a*sin(pointingDec));
    double xp = f*cos(delta)*sin(alpha - pointingRA);
    double xpr = xp;
    double ypr = yp;
    // for rotated telescope
    if(rotationAngle != 0) {
        xpr = xp*cos(rotationAngle) + yp*sin(rotationAngle);
        ypr = -xp*sin(rotationAngle) + yp*cos(rotationAngle);
    }
    // get chip offsets in the focal plane's reference frame
    double cx = xPos + deltaX - refX;
    double cy = yPos + deltaY - refY;
    double cz = deltaZ - refZ;
    double dx = xpr - cx;
    double dy = ypr - cy;
    if((phiChip != 0 || psiChip != 0 || thetaChip != 0) && (cx != 0 || cy != 0 || cz != 0)) {
        double a11 = cos(psiChip)*cos(phiChip) - cos(thetaChip)*sin(phiChip)*sin(psiChip);
        double a12 = cos(psiChip)*sin(phiChip) + cos(thetaChip)*cos(phiChip)*sin(psiChip);
        double a13 = sin(psiChip)*sin(thetaChip);
        double a21 = -sin(psiChip)*cos(phiChip) - cos(thetaChip)*sin(phiChip)*cos(psiChip);
        double a22 = -sin(psiChip)*sin(phiChip) + cos(thetaChip)*cos(phiChip)*cos(psiChip);
        double a23 = cos(psiChip)*sin(thetaChip);
        double cxr = a11*cx + a12*cy + a13*cz;
        double cyr = a21*cx + a22*cy + a23*cz;
        dx = xpr - cxr;
        dy = ypr - cyr;
    }
    // consider difference between aribtrary image plane orientation and arbitrary chip orientation
    if(phiImage != phiChip || psiImage != psiChip || thetaImage != thetaChip) {
        // chip rotation matrix
        double c11 = cos(psiChip)*cos(phiChip) - cos(thetaChip)*sin(phiChip)*sin(psiChip);
        double c12 = cos(psiChip)*sin(phiChip) + cos(thetaChip)*cos(phiChip)*sin(psiChip);
        double c21 = -sin(psiChip)*cos(phiChip) - cos(thetaChip)*sin(phiChip)*cos(psiChip);
        double c22 = -sin(psiChip)*sin(phiChip) + cos(thetaChip)*cos(phiChip)*cos(psiChip);
        double c31 = sin(thetaChip)*sin(phiChip);
        double c32 = -sin(thetaChip)*cos(phiChip);
        // image inverse rotation matrix (transpose matrix)
        double i11 = cos(psiImage)*cos(phiImage) - cos(thetaImage)*sin(phiImage)*sin(psiImage);
        double i12 = -sin(psiImage)*cos(phiImage) - cos(thetaImage)*sin(phiImage)*cos(psiImage);
        double i21 = cos(psiImage)*sin(phiImage) + cos(thetaImage)*cos(phiImage)*sin(psiImage);
        double i22 = -sin(psiImage)*sin(phiImage) + cos(thetaImage)*cos(phiImage)*cos(psiImage);
        double i31 = sin(psiImage)*sin(thetaImage);
        double i32 = cos(psiImage)*sin(thetaImage);
        // image inverse matrix times chip rotation matrix
        double r11 = i11*c11 + i21*c21 + i31*c31;
        double r12 = i12*c11 + i22*c21 + i32*c31;
        double r21 = i11*c12 + i21*c22 + i31*c32;
        double r22 = i12*c12 + i22*c22 + i32*c32;
        // apply rotation
        *x = r11*dx + r12*dy;
        *y = r21*dx + r22*dy;
    } else {
        *x = dx;
        *y = dy;
    }

}

void Trim::readCatalog() {

    gzFile gzfp[maxCatalog];
    FILE *fp[maxCatalog];
    FILE *fp2[maxChip];
    char line[4096];
    long sourceCount[maxChip];
    char ignore[2][4096];
    double ignore2;
    double ra, dec, mag;
    double x, y;
    double currentBuffer;


    for (int d = 0; d < nChip; d++) {
        sprintf(outputFilename[d], "trimcatalog_%s_%s.pars", obshistid.c_str(), chipid[d].c_str());
    }

    for (int d = 0; d < nChip; d++) {
        fp2[d] = fopen(outputFilename[d], "wt");
        if (fp2[d]==NULL) {
            std::cout << "Cannot open for writing output file " << outputFilename[d] << "\n.";
            exit(1);
        }
        fprintf(fp2[d], "\n");
        fprintf(fp2[d], "\n");
        sourceCount[d] = 0;
    }

    for (int c = 0; c < nCatalog; c++) {
        if (strstr(catalog[c].c_str(), ".gz")==NULL) {

            fp[c] = fopen(catalog[c].c_str(), "rt");
            if (fp[c] == NULL){
                std::cout << "Cannot open catalog file " << catalog[c] << "." << std::endl;;
                exit(1);
            }
            while (fgets( line, 4096, fp[c] )) {
                sscanf(line, "%s %lf %lf %lf %lf %s", ignore[0], &ignore2, &ra, &dec, &mag, ignore[1]);
                ra = ra*DEGREE;
                dec = dec*DEGREE;
                for (int d = 0; d < nChip; d++) {
                    currentBuffer = (buffer*pixelSize[d] + scale*pow(2.5, 17 - mag));
                    if (strayLight == 0 && currentBuffer > extendedBuffer + buffer*pixelSize[d]) currentBuffer = extendedBuffer + buffer*pixelSize[d];
                    xyPosition(ra, dec, phi[d], psi[d], theta[d], phiImage, psiImage, thetaImage, xPosition[d], yPosition[d], deltaX[d], deltaY[d], deltaZ[d], &x, &y);
                    if ((fabs(x) <= xDimension[d] + currentBuffer) &&
                        (fabs(y) <= yDimension[d] + currentBuffer)){
                        fprintf( fp2[d], "%s\n", line);
                        sourceCount[d]++;
                    } 
                }
            }
            fclose(fp[c]);

        } else {

            gzfp[c] = gzopen( catalog[c].c_str(), "r" );
            if (gzfp[c] == NULL) {
                std::cout <<"Cannot open catalog file " << catalog[c] << "." << std::endl;;
                exit(1);
            }
            while (gzgets( gzfp[c] , line, 4096)) {
                sscanf(line, "%s %lf %lf %lf %lf %s", ignore[0], &ignore2, &ra, &dec, &mag, ignore[1]);
                ra = ra*DEGREE;
                dec = dec*DEGREE;
                for (int d = 0; d < nChip; d++) {
                    currentBuffer = (buffer*pixelSize[d] + scale*pow(2.5, 17 - mag));
                    if (strayLight == 0 && currentBuffer > extendedBuffer + buffer*pixelSize[d]) currentBuffer = extendedBuffer + buffer*pixelSize[d];
                    xyPosition(ra, dec, phi[d], psi[d], theta[d], phiImage, psiImage, thetaImage, xPosition[d], yPosition[d], deltaX[d], deltaY[d], deltaZ[d], &x, &y);
                    if ((fabs(x) <= xDimension[d] + currentBuffer) &&
                        (fabs(y) <= yDimension[d] + currentBuffer)){
                        fprintf(fp2[d], "%s\n", line);
                        sourceCount[d]++;
                    }
                }
            }
            gzclose(gzfp[c]);

        }
    }

    for (int d = 0; d < nChip; d++) {
        fclose(fp2[d]);
    }

    long totalSourceCount = 0;
    for (int d = 0; d < nChip; d++) {
        totalSourceCount += sourceCount[d];
    }
    if (totalSourceCount>=minSource) {
        std::cout << "------------------------------------------------------------------------------------------" << std::endl;;
        std::cout << "Trim Catalog" << std::endl;;
        std::cout << "------------------------------------------------------------------------------------------" << std::endl;;
        for (int d = 0; d < nChip; d++) {
            if (sourceCount[d] >= minSource) {
                std::cout << "Found " << sourceCount[d] << " source(s) for chip " << chipid[d] << "." << std::endl;;
            }
        }
    }

}

void Trim::getDetectorProperties( int d ){

    std::istringstream focalplanePars(readText::get(instrdir + "/focalplanelayout.txt", chipid[d]));
    focalplanePars >> xPosition[d] >> yPosition[d] >> pixelSize[d] >> xDimension[d] >> yDimension[d];
    xDimension[d] *= pixelSize[d]/2.0;
    yDimension[d] *= pixelSize[d]/2.0;
    std::string tempstring;
    double temp;
    focalplanePars >> tempstring;
    focalplanePars >> temp;
    focalplanePars >> temp;
    focalplanePars >> tempstring;
    focalplanePars >> phi[d];
    focalplanePars >> psi[d];
    focalplanePars >> theta[d];
    phi[d] *= DEGREE;
    psi[d] *= DEGREE;
    theta[d] *= DEGREE;
    focalplanePars >> deltaX[d];
    focalplanePars >> deltaY[d];
    focalplanePars >> deltaZ[d];
    deltaX[d] *= 1000.0;
    deltaY[d] *= 1000.0;
    deltaZ[d] *= 1000.0;

}

void Trim::setup() {


    instrdir = "../data/lsst";
    flatDirectory = 0;

    readText pars(std::cin);


    for (size_t t(0); t < pars.getSize(); t++) {
        std::string line(pars[t]);
        readText::get(line, "pointingra", pointingRA);
        readText::get(line, "pointingdec", pointingDec);
        readText::get(line, "rotationangle", rotationAngle);
        readText::get(line, "filter", filter);
        readText::get(line, "obshistid", obshistid);
        readText::get(line, "instrdir", instrdir);
        readText::get(line, "flatdir", flatDirectory);
        readText::get(line, "minsource", minSource);
        readText::get(line, "chipid", chipid);
        readText::get(line, "catalog", catalog);
    }
    pointingRA *= DEGREE;
    pointingDec *= DEGREE;
    rotationAngle *= DEGREE;

    nCatalog = catalog.size();
    nChip = chipid.size();
    if (flatDirectory == 1) instrdir = ".";

    std::istringstream wavelengthPars(readText::get(instrdir + "/central_wavelengths.txt", filter));
    double wavelength;
    wavelengthPars >> wavelength >> plateScale >> refX >> refY >> refZ >> phiImage >> psiImage >> thetaImage;
    phiImage *= DEGREE;
    psiImage *= DEGREE;
    thetaImage *= DEGREE;
    buffer = 100;
    strayLight = 1;
    focalLength = plateScale/DEGREE;
    scale = (0.2*ARCSEC/DEGREE)*(focalLength*DEGREE);
    extendedBuffer = (60.0*ARCSEC/DEGREE)*(focalLength*DEGREE);

    if (nChip > maxChip || nCatalog > maxCatalog ){
        std::cout << "Cannot split into that many files." << std::endl;
        exit(1);
    }

}
